@@ -27,6 +27,13 @@
 #define PPP_IP_PROTO	0x0021
 #define MD5		5
 
+#define DBG(p, fmt, arg...) do {				\
+	char *str = g_strdup_printf("%s:%s() " fmt, __FILE__,	\
+					__FUNCTION__ , ## arg); \
+	ppp_debug(p, str);					\
+	g_free(str);						\
+} while (0)
+
 struct ppp_chap;
 struct ppp_net;
 
@@ -95,11 +102,13 @@ void ppp_chap_free(struct ppp_chap *chap);
 void ppp_chap_process_packet(struct ppp_chap *chap, const guint8 *new_packet);
 
 /* TUN / Network related functions */
-struct ppp_net *ppp_net_new(GAtPPP *ppp);
+struct ppp_net *ppp_net_new(GAtPPP *ppp, int fd);
 const char *ppp_net_get_interface(struct ppp_net *net);
 void ppp_net_process_packet(struct ppp_net *net, const guint8 *packet);
 void ppp_net_free(struct ppp_net *net);
 gboolean ppp_net_set_mtu(struct ppp_net *net, guint16 mtu);
+void ppp_net_suspend_interface(struct ppp_net *net);
+void ppp_net_resume_interface(struct ppp_net *net);
 
 /* PPP functions related to main GAtPPP object */
 void ppp_debug(GAtPPP *ppp, const char *str);
