@@ -116,39 +116,44 @@ static void add_mbm(struct ofono_modem *modem,
 	DBG("desc: %s", desc);
 
 	registered = ofono_modem_get_integer(modem, "Registered");
-	if (registered != 0)
-		return;
 
-	if (g_str_has_suffix(desc, "Minicard Modem") ||
+	if (registered == 0 &&
+			(g_str_has_suffix(desc, "Minicard Modem") ||
 			g_str_has_suffix(desc, "Minicard Modem 2") ||
 			g_str_has_suffix(desc, "Mini-Card Modem") ||
 			g_str_has_suffix(desc, "Broadband Modem") ||
-			g_str_has_suffix(desc, "Broadband USB Modem")) {
+			g_str_has_suffix(desc, "Broadband USB Modem"))) {
 		devnode = udev_device_get_devnode(udev_device);
+
 		if (ofono_modem_get_string(modem, MODEM_DEVICE) == NULL)
 			ofono_modem_set_string(modem, MODEM_DEVICE, devnode);
 		else
 			ofono_modem_set_string(modem, DATA_DEVICE, devnode);
-	} else if (g_str_has_suffix(desc, "Minicard Data Modem") ||
+	} else if (registered == 0 &&
+			(g_str_has_suffix(desc, "Minicard Data Modem") ||
 			g_str_has_suffix(desc, "Mini-Card Data Modem") ||
-			g_str_has_suffix(desc, "Broadband Data Modem")) {
+			g_str_has_suffix(desc, "Broadband Data Modem"))) {
 		devnode = udev_device_get_devnode(udev_device);
 		ofono_modem_set_string(modem, DATA_DEVICE, devnode);
 	} else if (g_str_has_suffix(desc, "Minicard GPS Port") ||
-			g_str_has_suffix(desc, "Mini-Card GPRS Port") ||
+			g_str_has_suffix(desc, "Mini-Card GPS Port") ||
 			g_str_has_suffix(desc, "Broadband GPS Port")) {
 		devnode = udev_device_get_devnode(udev_device);
 		ofono_modem_set_string(modem, GPS_DEVICE, devnode);
-	} else if (g_str_has_suffix(desc, "Minicard Network Adapter") ||
+	} else if (registered == 0 &&
+			(g_str_has_suffix(desc, "Minicard Network Adapter") ||
 			g_str_has_suffix(desc, "Mini-Card Network Adapter") ||
 			g_str_has_suffix(desc, "Broadband Network Adapter") ||
-			g_str_has_suffix(desc, "Minicard NetworkAdapter")) {
+			g_str_has_suffix(desc, "Minicard NetworkAdapter"))) {
 		devnode = get_property(udev_device, "INTERFACE");
 		ofono_modem_set_string(modem, NETWORK_INTERFACE, devnode);
 	} else {
 		return;
 	}
 
+	if (registered == 1)
+		return;
+
 	device  = ofono_modem_get_string(modem, MODEM_DEVICE);
 	data = ofono_modem_get_string(modem, DATA_DEVICE);
 	network = ofono_modem_get_string(modem, NETWORK_INTERFACE);
@@ -552,6 +557,36 @@ static void add_nokiacdma(struct ofono_modem *modem,
 	ofono_modem_register(modem);
 }
 
+static void add_linktop(struct ofono_modem *modem,
+					struct udev_device *udev_device)
+{
+	const char *devnode, *intfnum;
+	struct udev_device *parent;
+	int registered;
+
+	DBG("modem %p", modem);
+
+	registered = ofono_modem_get_integer(modem, "Registered");
+	if (registered != 0)
+		return;
+
+	parent = udev_device_get_parent(udev_device);
+	intfnum = udev_device_get_sysattr_value(parent, "bInterfaceNumber");
+
+	DBG("intfnum %s", intfnum);
+
+	if (g_strcmp0(intfnum, "01") == 0) {
+		devnode = udev_device_get_devnode(udev_device);
+		ofono_modem_set_string(modem, "Modem", devnode);
+	} else if (g_strcmp0(intfnum, "03") == 0) {
+		devnode = udev_device_get_devnode(udev_device);
+		ofono_modem_set_string(modem, "Control", devnode);
+
+		ofono_modem_set_integer(modem, "Registered", 1);
+		ofono_modem_register(modem);
+	}
+}
+
 static void add_modem(struct udev_device *udev_device)
 {
 	struct ofono_modem *modem;
@@ -636,6 +671,8 @@ done:
 		add_nokia(modem, udev_device);
 	else if (g_strcmp0(driver, "isiusb") == 0)
 		add_isi(modem, udev_device);
+	else if (g_strcmp0(driver, "u8500") == 0)
+		add_isi(modem, udev_device);
 	else if (g_strcmp0(driver, "n900") == 0)
 		add_isi(modem, udev_device);
 	else if (g_strcmp0(driver, "gobi") == 0)
@@ -646,6 +683,8 @@ done:
 		add_tc65(modem, udev_device);
 	else if (g_strcmp0(driver, "nokiacdma") == 0)
 		add_nokiacdma(modem, udev_device);
+        else if (g_strcmp0(driver, "linktop") == 0)
+		add_linktop(modem, udev_device);
 }
 
 static gboolean devpath_remove(gpointer key, gpointer value, gpointer user_data)
